# -*- coding:utf-8 -*-

import re

from types import BooleanType, FloatType, IntType, LongType, NoneType, StringType, UnicodeType
from types import ListType, DictType

from grapheekdb.lib.exceptions import GrapheekInvalidDataTypeException


VALID_KEY = re.compile('^[a-zA-A][a-zA-Z0-9_]*$')


def check_valid_value(value):
    if isinstance(value, (BooleanType, FloatType, IntType, LongType, NoneType, StringType, UnicodeType)):
        return True
    elif isinstance(value, ListType):
        return all([check_valid_data(subvalue) for subvalue in value])
    elif isinstance(value, DictType):
        return check_valid_data(value, check_dict=False)
    return False


def check_valid_data(data, check_dict=True):
    # Checking that data is a dictionnary
    if check_dict and not isinstance(data, DictType):
        raise GrapheekInvalidDataTypeException('data must be a dictionnary')
    # Ckecking that data keys match '^[a-zA-A][a-zA-Z0-9]*$' regular expression
    if not(all(VALID_KEY.match(key) for key in data)):
        raise GrapheekInvalidDataTypeException('a data key must only contain letter and digit and start with a letter')
    # Ckecking that data keys don't contain a double underscore (could mess lookup)
    if any('__' in key for key in data):
        raise GrapheekInvalidDataTypeException('a data key must not contain a double underscore')
    # Ckecking that data keys don't end with an underscore (could mess lookup)
    if any(key.endswith('_') for key in data):
        raise GrapheekInvalidDataTypeException('a data key must not end with an underscore')
    # Values validations  :
    if not all([check_valid_value(value) for value in data.values()]):
        raise GrapheekInvalidDataTypeException('value must be one of following type : boolean/float/integer/long/none/string/unicode/dict/list')
    return True



# -*- coding:utf-8 -*-

from functools import partial

import lookups

from grapheekdb.lib.undef import UNDEFINED

from grapheekdb.backends.data.keys import build_key
from grapheekdb.backends.data.keys import DATA_SUFFIX

from grapheekdb.lib.exceptions import GrapheekSubLookupNotImplementedException
from grapheekdb.lib.exceptions import GrapheekInvalidLookupException


def _get_exact_filters(**filters):
    exact_filters = {}
    for key, value in filters.items():
        # Note : not only looking for exact lookups
        lst = key.split('__')
        len_lst = len(lst)
        if len_lst == 1:
            field = lst[0]
            clause = 'exact'
            exact_filters[field] = value
        elif len_lst == 2:
            field, clause = lst
            if clause == 'exact':
                exact_filters[field] = value
        else:
            raise GrapheekSubLookupNotImplementedException
    return exact_filters


def build_attr_filter_funcs(**filters):
    field = clause = None
    attr_filter_funcs = []
    for key, value in filters.items():
        lst = key.split('__')
        len_lst = len(lst)
        if len_lst == 1:
            field = lst[0]
            clause = 'exact'
            lookup_func = lookups.lookup_exact
        elif len_lst == 2:
            field, clause = lst
            try:
                lookup_func = getattr(lookups, 'lookup_' + lst[1])
            except AttributeError:
                raise GrapheekInvalidLookupException
        else:
            raise GrapheekSubLookupNotImplementedException
        attr_filter_funcs.append((field, clause, partial(lookup_func, value)))
    return attr_filter_funcs


def filter_entities(graph, kind, iterator, filter_funcs, on_item=False):
    for entity_id in iterator:
        if filter_funcs:
            data = graph._get(None, build_key(kind, entity_id, DATA_SUFFIX))
            if data == UNDEFINED:  # pragma : no cover
                # This could happen because of concurrent modifications (my different process or different threads)
                # This won't normally occur when load testing with greenlets -> thus no cover
                # (or worst : direct access to backend by another "process")
                continue
            valid = True
            for field, clause, filter_func in filter_funcs:
                field_value = data.get(field, UNDEFINED)
                if field_value == UNDEFINED:
                    if clause == 'isnull':
                        field_value = None
                    else:
                        valid = False
                        break
                try:
                    valid = filter_func(field_value)
                except:
                    # Whatever happens, it shouldn't
                    # So, let's consider that data wasn't ok for the filter
                    valid = False
                if not(valid):
                    break
            if valid:
                if on_item:
                    yield graph._item_from_id(kind, entity_id)
                else:
                    yield entity_id
        else:
            # I don't do : yield item if on_item else entity_id
            # Because item is not computed (when on_item is False, this is not useful to instantiate an Entity)
            if on_item:
                yield graph._item_from_id(kind, entity_id)
            else:
                yield entity_id
